/* implement a hash table
	94/01/06 aih - allocation of items is left to the application
	93/12/18 aih - uses memory manager to allocate memory
					 - added function to rehash table
	93/10/31 aih - uses calloc/free rather than memory manager
	93/10/28 aih - created
*/

#include <math.h>
#include <time.h>
#include "HashLib.h"
#include "MemoryLib.h"
#include "SortLib.h"

/* true if the hash table is valid */
Boolean HashValid(HashPtr hash)
{
	if (! hash) return(false);
	if (! hash->table) return(false);
	if (hash->size < HASH_MINSIZE) return(false);
	if (hash->size % HASH_MINSIZE == 0) return(false);
	return(true);
}

/* return a new hash table */
HashPtr HashBegin(size_t size)
{
	volatile HashPtr hash = NULL;
	
	require(size >= HASH_MINSIZE);
	require(size % HASH_MINSIZE != 0);
	TRY {
		hash = PtrBeginClear(sizeof(HashType));
		hash->table = PtrBeginClear(size * sizeof(LLType));
		hash->size = size;
	} CATCH {
		HashEnd(hash, NULL);
	} ENDTRY;
	ensure(HashValid(hash));
	return(hash);
}

/* dispose of the hash table; the dispose function is called for each item */
void HashEnd(HashPtr hash, HashDisposeType dispose)
{
	size_t i;

	require(! hash || HashValid(hash));
	if (hash) {
		if (dispose) {
			for (i = 0; i < hash->size; i++) {
				while (hash->table[i]) {
					LLPtr item = hash->table[i];
					hash->table[i] = LLPDelete(hash->table[i], item);
					dispose(item);
				}
			}
		}
		PtrEnd(hash->table);
		PtrEnd(hash);
		hash = NULL;
	}
	ensure(! HashValid(hash));
}

/* Return the first prime number larger than 'minsize', abort if no
	prime was found in 'maxtime' clock cycles and return first odd
	integer greater than or equal to 'minsize'. Use this function to
	calculate a size for a hash table. */
size_t HashCalculateSize(size_t minsize, clock_t maxtime)
{
	register size_t i, n, p;
	long start;
	
	require(minsize >= HASH_MINSIZE);
	if (minsize % 2 == 0)
		minsize++;
	i = n = 0;
	start = TickCount();
	for (p = minsize; TickCount() - start < maxtime; p += 2) {
		n = sqrt(p);
		for (i = 3; i <= n && p % i != 0; i += 2)
			;
		if (i > n)
			break;
	}
	if (i <= n)
		p = minsize;
	ensure(p >= minsize && p % 2 != 0);
	return(p);
}

/* return size of hash table */
size_t HashSize(HashPtr hash)
{
	require(HashValid(hash));
	return(hash->size);
}

/* return number of items in hash table */
size_t HashNItems(HashPtr hash)
{
	require(HashValid(hash));
	return(hash->nitems);
}

/* return the i'th entry in the hash table */
LLPtr HashEntry(HashPtr hash, size_t index)
{
	require(HashValid(hash));
	require(0 <= index && index < hash->size);
	return(hash->table[index]);
}

/* calculate the index into the hash table for the key */
size_t HashIndex(HashPtr hash, const void *key, register size_t n)
{
	register const unsigned char *p = key;
	register size_t index;
	register size_t size;
	
	require(HashValid(hash));
	require(n > 0);
	size = hash->size;
	index = *p++; n--;
	while (n-- > 0)
		index = ((index << CHAR_BIT) + *p++) % size;
	ensure(0 <= index && index < size);
	return(index);
}

/* return item matching data */
static LLPtr LLPFind(LLPtr head, const void *key, HashCompareType compare)
{
	while (head) {
		int cmp = compare(key, head);
		if (cmp == 0)
			break;
		head = LLPNext(head);
	}
	return(head);
}

/* find key in table, return entry if found, else NULL */
LLPtr HashFind(HashPtr hash, const void *key, size_t index,
	HashCompareType compare)
{
	return(LLPFind(HashEntry(hash, index), key, compare));
}

/* insert an item into the hash table */
void HashInsert(HashPtr hash, LLPtr item, size_t index)
{
	require(HashValid(hash));
	require(0 <= index && index < hash->size);
	hash->table[index] = LLPInsert(hash->table[index], item);
	hash->nitems++;
	ensure(HashValid(hash));
}

/* delete an item from the hash table */
void HashDelete(HashPtr hash, LLPtr item, size_t index)
{
	require(HashValid(hash));
	require(hash->nitems > 0);
	require(0 <= index && index < hash->size);
	hash->table[index] = LLPDelete(hash->table[index], item);
	hash->nitems--;
	ensure(HashValid(hash));
}

/* Return an array containing pointers to the hash table's items in sorted
	order. The comparison function is called via qksort (which is
	equivalent to qsort). Dispose of the table by calling PtrEnd. */
LLPtr *HashSort(HashPtr hash, HashSortType compare)
{
	LLPtr *table = NULL;
	LLPtr item;
	size_t i, j;
	
	require(HashValid(hash));
	table = PtrBegin(hash->nitems * sizeof(LLPtr));
	j = 0;
	for (i = 0; i < hash->size; i++) {
		for (item = hash->table[i]; item; item = LLPNext(item)) {
			check(j < hash->nitems);
			table[j++] = item;
		}
	}
	check(j == hash->nitems);
	qksort(table, hash->nitems, sizeof(LLPtr), compare);
	return(table);
}

#ifdef NOT_READY

/* rebuild the hash table, return new hash table */
HashPtr HashRehash(HashPtr hash, size_t size)
{
	HashPtr rehash = NULL;
	LLPtr item = NULL;
	size_t nitems = 0;
	size_t i = 0;

	require(HashValid(hash));
	TRY {
		nitems = HashNItems(hash);
		rehash = HashBegin(size);
		for (i = 0; i < hash->size; i++) {
			for (item = hash->table[i]; item; item = LListNext(item))
				do_insert(rehash, item, HashIndex(rehash, item));
		}
		HashEnd(hash, NULL);
	} CATCH {
		HashEnd(rehash, NULL);
	} ENDTRY;
	ensure(HashNItems(rehash) == nitems);
	return(rehash);
}

#endif /* NOT_READY */