/*	General purpose binary tree library supporting splay trees. 
		
	Revision history:

	94/01/06 aih
	- Allocation and disposal of nodes is left to the caller. This allows
	more efficient allocation for specific applications, rather than using
	the memory manager or the standard library.
	
	93/10/29 aih
	- Converted to use calloc and free, since using the memory manager's
	handles was just the slowest thing--the memory manager is designed to
	infrequently allocate large blocks, not to allocate many small blocks.
	THINK C's implementation of the standard C allocators was many times
	faster. This also made the code easier to read, replacing ugly
	double derefrences with a single dereference.
	
	93/10/28 aih
	- simplified interface
	
	93/10/25 aih
	- updated for newer version of libraries
	- added node creation functions
	- added disposal of all nodes in a tree when the tree is disposed of
	
	91/07/02 AIH
	- Added some assertions and comments
	
	91/04/18 AIH
	- Added splay tree functions
	
	91/04/09 AIH
	- Exported TreeValid
	
	91/03/? Ari Halberstadt (AIH)
	- Completely revised and simplified library sometime in late March, 1991. */

#include <stdlib.h>
#include <string.h>
#include "MemoryLib.h"
#include "TreeLib.h"

/*----------------------------------------------------------------------------*/
/* debugging */
/*----------------------------------------------------------------------------*/

#ifndef NDEBUG

#include <stdio.h>

/* print a node */
static void NodePrint(NodePtr node, const char *prompt)
{
	if (node) {
		printf("%s: 0x%lx, parent=0x%lx, left=0x%lx, right=0x%lx\n",
			prompt, node, node->parent, node->left, node->right);
	}
}

#endif /* NDEBUG */


/*----------------------------------------------------------------------------*/
/* operations on nodes */
/*----------------------------------------------------------------------------*/

/* true if node is valid */
static Boolean NodeValid(NodePtr node)
{
	if (! node) return(false);
	if (node->parent &&
			(node->parent->left != node &&
			 node->parent->right != node)) return(false);
	if (node->left && node->left->parent != node) return(false);
	if (node->right && node->right->parent != node) return(false);
	return(true);
}

/*----------------------------------------------------------------------------*/
/* operations on trees */
/*----------------------------------------------------------------------------*/

/* true if the tree is valid */
Boolean TreeValid(TreePtr tree)
{
	if (! tree) return(false);
	if (! tree->root && tree->size) return(false);
	if (! tree->size && tree->root) return(false);
	if (tree->root && ! NodeValid(tree->root)) return(false);
	if (tree->root && tree->root->parent) return(false);
	return(true);
}

/* create a new tree */
TreePtr TreeBegin(Boolean splay)
{
	TreePtr tree;
	
	tree = PtrBeginClear(sizeof(TreeType));
	tree->splay = splay;
	ensure(TreeValid(tree));
	return(tree);
}

/* dispose of the tree; the dispose function is called for each node */
void TreeEnd(TreePtr tree, TreeVisitType dispose)
{
	if (tree) {
		if (dispose && TreeRoot(tree))
			TreePostOrder(TreeRoot(tree), dispose, NULL);
		PtrEnd(tree);
	}
}

/* return number of nodes in tree */
size_t TreeSize(TreePtr tree)
{
	require(TreeValid(tree));
	return(tree->size);
}

/* return root node of tree */
NodePtr TreeRoot(TreePtr tree)
{
	require(TreeValid(tree));
	return(tree->root);
}

/*----------------------------------------------------------------------------*/
/* Splay tree functions based on:

	Sleator, Daniel Dominic and Tarjan, Robert Endre "Self-Adjusting
		Binary Search Trees", JACM, 1985, p652-686, and the implementation
		given on p666-667.
	Williams, Kent, "Smart Searching with C++", Computer Language, September
	1990, p45-56. */
/*----------------------------------------------------------------------------*/

#define left(x)	x->left
#define right(x)	x->right
#define l(x)		x->left
#define r(x)		x->right
#define p(x)		x->parent
#define g(x)		p(p(x))

/* rotate the edge between a node and its right child */
static void RotateLeft(NodePtr y)
{
	NodePtr x = right(y), z = p(y);

	require(NodeValid(y));
	if (z) {
		if (left(z) == y)
			left(z) = x;
		else
			right(z) = x;
	}
	if (x) {
		right(y) = left(x);
		left(x) = y;
		p(x) = z;
	}
	p(y) = x;
	if (right(y))
		p(right(y)) = y;
	ensure(NodeValid(y));
}

/* rotate the edge between a node and its left child */
static void RotateRight(NodePtr y)
{
	NodePtr x = left(y), z = p(y);

	require(NodeValid(y));
	if (z) {
		if (left(z) == y)
			left(z) = x;
		else
			right(z) = x;
	}
	if (x) {
		left(y) = right(x);
		right(x) = y;
		p(x) = z;
	}
	p(y) = x;
	if (left(y))
		p(left(y)) = y;
	ensure(NodeValid(y));
}

/* Splay the tree at the node indicated by the cursor. As recommended in
	Sleator and Tarjan, for greater efficiency I've expanded the rotations
	in-line and removed redundant tests and assignments. The optimizations
	also make liberal use of compact C code, which may make the code a little
	obscure, though I've also tried to make liberal use of comments. */
void TreeSplay(TreePtr tree, register NodePtr x)
{
	register NodePtr y, z;
	
	require(TreeValid(tree));
	require(! x || NodeValid(x));
	if (! x || ! tree->splay)
		return;
	while (y = p(x)) {
		z = p(y);
		if (p(x) = (z ? p(z) : NULL)) {		/* make x a child of the parent of z */
			if (l(p(z)) == z) l(p(z)) = x;
			else r(p(z)) = x;
		}
		if (x == l(y)) { 
			if (! z) {								/* zig */
				if (l(y) = r(x)) p(l(y)) = y;	/* make the right child of x the left child of y */
				r(x) = y; p(y) = x;				/* make y the right child of x */
			}
			else if (y == l(z)) {				/* zig-zig */
				if (l(y) = r(x)) p(l(y)) = y;	/* make the right child of x the left child of y */
				if (l(z) = r(y)) p(l(z)) = z;	/* make the right child of y the left child of z */
				r(x) = y; p(y) = x;				/* make y the right child of x */
				r(y) = z; p(z) = y;				/* make z the right child of y */
			}
			else {									/* zig-zag */
				if (r(z) = l(x)) p(r(z)) = z;	/* make left child of x the right child of z */
				if (l(y) = r(x)) p(l(y)) = y;	/* make right child of x the left child of y */
				l(x) = z; p(z) = x;				/* make z the left child of x */
				r(x) = y; p(y) = x;				/* make y the right child of x */
			}
		}
		else {
			if (! z) {								/* zig */
				if (r(y) = l(x)) p(r(y)) = y;	/* make the left child of x the right child of y */
				l(x) = y; p(y) = x;				/* make y the left child of x */
			}
			else if (y == r(z)) {				/* zig-zig */
				if (r(y) = l(x)) p(r(y)) = y;	/* make the left child of x the right child of y */
				if (r(z) = l(y)) p(r(z)) = z;	/* make the left child of y the right child of z */
				l(x) = y; p(y) = x;				/* make y the left child of x */
				l(y) = z; p(z) = y;				/* make z the left child of y */
			}
			else {									/* zig-zag */
				if (l(z) = r(x)) p(l(z)) = z;	/* make right child of x the left child of z */
				if (r(y) = l(x)) p(r(y)) = y;	/* make left child of x the right child of y */
				r(x) = z; p(z) = x;				/* make z the right child of x */
				l(x) = y; p(y) = x;				/* make y the left child of x */
			}
		}
	}
	tree->root = x;
	ensure(TreeValid(tree));
}

/*----------------------------------------------------------------------------*/
/* searching, inserting, and deleting */
/*----------------------------------------------------------------------------*/

/*
From Sleator and Tarjan (p660-661):

acces(i, t):	If item i is in tree t, return a pointer to its location;
					otherwise, return a pointer to the null node.
insert(i, t):	Insert item i in tree t, assuming that it is not there already.
delete(i, t):	Delete item i from tree t, assuming that it is present.
join(t1, t2):	Combine trees t1 and t2 into a single tree containing all
					items from both trees and return the resulting tree.
					This operation assumes that all items in t1 are less than
					all those in t2 and destroys both t1 and t2.
split(i, t):	Construct and return two trees t1 and t2, where t1 contains
					all items in t less than or equal to i, and t2 contains all
					items in t greater than i. This operation destroys t.
*/
					
/* join the subtrees, return new root, assumes left < right */
static NodePtr join(NodePtr left, NodePtr right)
{
	NodePtr join;
	
	require(! left || NodeValid(left));
	require(! right || NodeValid(right));
	/* require(largest(left) < smallest(right)); */
	if (! right)
		join = left;
	else if (! left)
		join = right;
	else {
		/* move left subtree to left subtree of prior parent's successor,
			making it just like the first case, where right == NULL */
		NodePtr successor = right;
		while (successor->left)
			successor = successor->left;
		successor->left = left;
		left->parent = successor;
		join = right;
	}
	ensure(! join || NodeValid(join));
	return(join);
}

/* search for key in tree and splay at node or parent if not found */
NodePtr TreeAccess(TreePtr tree, const void *key, TreeCompareType compare)
{
	NodePtr parent;
	NodePtr node;
		
	node = TreeFindAt(tree, key, compare, &parent, NULL);
	TreeSplay(tree, node ? node : parent);
	return(node);
}

/* Return node matching item, parent of last node accessed, and result
	of last comparison. The tree is not splayed. Used for insertion
	into a tree. */
NodePtr TreeFindAt(TreePtr tree, register const void *key,
	register TreeCompareType compare, NodePtr *pparent, int *pcmp)
{
	register NodePtr parent = NULL;
	register NodePtr node = NULL;
	register int cmp = -1;
	
	require(TreeValid(tree));
	node = tree->root;
	while (node) {
		parent = node;
		cmp = compare(key, node);
		if (cmp < 0)
			node = node->left;
		else if (cmp > 0)
			node = node->right;
		else
			break;
	}
	if (pcmp) *pcmp = cmp;
	if (pparent) *pparent = parent;
	ensure(! tree->root ? (! node && ! parent) : NodeValid(parent));
	ensure(! node ? cmp != 0 : (NodeValid(node) && cmp == 0 && parent == node));
	return(node);
}

/* insert node as a child of parent */
void TreeInsertAt(TreePtr tree, NodePtr node, NodePtr parent, int cmp)
{
	require(TreeValid(tree));
	check(cmp != 0);
	check(! parent || NodeValid(parent));
	if (parent) {
		check(cmp > 0 ? ! parent->right : ! parent->left);
		if (cmp > 0)
			parent->right = node;
		else
			parent->left = node;
		node->parent = parent;
	}
	else {
		check(! tree->root);
		tree->root = node;
	}
	tree->size++;
	TreeSplay(tree, node);
}

/* insert item into tree */
void TreeInsert(TreePtr tree, NodePtr newnode, const void *key,
	TreeCompareType compare)
{
	int cmp = 0;
	NodePtr parent = NULL;	
	NodePtr node = NULL;

	node = TreeFindAt(tree, key, compare, &parent, &cmp);
	check(! node);
	TreeInsertAt(tree, newnode, parent, cmp);
}

/* delete node from tree */
void TreeDelete(TreePtr tree, NodePtr node)
{
	/* From Sleator and Tarjan p662: "To carry out delete(i, t), we search for
		the node containing i. Let this node be x and let its parent be
		y. We replace x as a child of y by the join of the left and right
		subtrees of x, and then we splay at y." */
	require(TreeValid(tree));
	require(tree->size > 0);
	if (node) {
		NodePtr joined = join(node->left, node->right);
		if (node->parent) {
			if (node->parent->left == node)
				node->parent->left = joined;
			else
				node->parent->right = joined;
		}
		else {
			check(node == tree->root);
			tree->root = joined;
		}
		if (joined)
			joined->parent = node->parent;
		tree->size--;
		TreeSplay(tree, node->parent);
	}
}

/*----------------------------------------------------------------------------*/
/* Traversing the tree. These functions are naive implementations and are
	rather inefficient in their use of stack space. Since every node includes
	a pointer to its parent it is possible to rewrite these as iterative
	functions, but the code would be more complicated and tricky to get right. */
/*----------------------------------------------------------------------------*/

/* traverse tree in in-order (left subtree, node, right subtree) */
void TreeInOrder(NodePtr node, TreeVisitType visit, void *data)
{
	require(NodeValid(node));
	if (node->left) TreeInOrder(node->left, visit, data);
	visit(node, data);
	if (node->right) TreeInOrder(node->right, visit, data);
}

/* traverse tree in pre-order (node, left subtree, right subtree) */
void TreePreOrder(NodePtr node, TreeVisitType visit, void *data)
{
	require(NodeValid(node));
	visit(node, data);
	if (node->left) TreePreOrder(node->left, visit, data);
	if (node->right) TreePreOrder(node->right, visit, data);
}

/* traverse tree in post-order (left subtree, right subtree, node) */
void TreePostOrder(NodePtr node, TreeVisitType visit, void *data)
{	
	require(NodeValid(node));
	if (node->left) TreePostOrder(node->left, visit, data);
	if (node->right) TreePostOrder(node->right, visit, data);
	visit(node, data);
}

static size_t max(size_t a, size_t b)
{
	return(a >= b ? a : b);
}

/* return height of node */
size_t TreeHeight(NodePtr node, size_t height)
{
	for (;;) {
		if (node->left && node->right) {
			return(max(TreeHeight(node->left, height + 1),
						  TreeHeight(node->right, height + 1)));
		}
		else if (node->left) {
			node = node->left;
			height++;
		}
		else if (node->right) {
			node = node->right;
			height++;
		}
		else
			return(height);
	}
}
